from functools import cached_property, lru_cache

import numpy as np
import matplotlib.pyplot as plt
from scipy.stats import linregress as linreg
from scipy.optimize import curve_fit
from scipy import constants as const

from tools.data import DatapointType

def gate_capacitance(rel, thick):
	# Returns in units F/m^2
	return const.epsilon_0 * rel / thick

def mobility(cox, gate, cond, strat="linreg", err=None):
	if strat == "linreg":
		a, b, r, _, stderr = linreg(gate, cond)
	elif strat == "curve_fit":
		model = lambda x, a, b: a*x+b

		# Get initial guess for parameters using linreg
		param = mobility(cox, gate, cond)

		popt, pcov = curve_fit(model, gate, cond, param[3:], sigma=err, absolute_sigma=True)
		all_err = np.sqrt(np.diag(pcov))

		a, b = popt
		stderr = all_err[0]
		r = 0

	mob = 1/cox * np.abs(a)
	stderr *= 1/cox * stderr / np.abs(stderr)

	return mob, r, stderr, a, b

def fit_to_units(*args):
	# Convert args to a list so that we can manipulate it directly
	args = list(args)

	# idx 3 and 4 are the slope and intercept if given len = 5, else assume 0 and 1
	idx = (0, 1)
	if len(args) == 5:
		idx = (3, 4)

	for i in idx:
		args[i] *= const.h / const.e**2

	return args

def get_direction_idx(bias):
	idx_f = np.where(np.roll(bias, -1) >= bias)[0]
	idx_b = np.where(np.roll(bias, -1) < bias)[0]

	# I have no idea why, but computing the reverse indices this way doesn´t work
	#idx_b = np.arange(0, len(bias))[~idx_f]

	if len(bias) != len(idx_f) + len(idx_b):
		raise ArithmeticError(f"Length of input and output not identical: {len(bias)} in, {len(idx_f)} + {len(idx_b)} = {len(idx_f) + len(idx_b)} out")

	return idx_f, idx_b

def fullplot(bias, r, sort=None, uniq=None, suffix=""):
	def plotax(ax_r, ax_c, x, y_r, y_c, marker='.', label=""):
		ax_r[0].plot(x[0], y_r[0], '.', label=label)
		ax_r[1].plot(x[1], y_r[1], '.', label=label)
		
		ax_c[0].plot(x[0], y_c[0], '.', label=label)
		ax_c[1].plot(x[1], y_c[1], '.', label=label)

	plt.figure(figsize=(16,16))
	ax_r = [plt.subplot(2, 2, i) for i in (1,2)]
	ax_c = [plt.subplot(2, 2, i) for i in (3,4)]
	
	idx_f, idx_b = get_direction_idx(bias)
	
	c = 1/r
	cu = c * const.h / const.e**2
	
	if uniq is None:
		plotax(ax_r, ax_c, 
			(bias[idx_f], bias[idx_b]),
			(r[idx_f], r[idx_b]),
			(cu[idx_f], cu[idx_b]),
			label=f"Measurement {suffix}"
		)

	else:
		for u in uniq:
			idx_u = np.where(sort == u)[0]
			u_idx_f = np.intersect1d(idx_f, idx_u)
			u_idx_b = np.intersect1d(idx_b, idx_u)
			
			if len(idx_u) != len(u_idx_f) + len(u_idx_b):
				raise ArithmeticError(
					f"Sorting lengths not equal: {len(idx_u)} in, " \
					+ f"{len(u_idx_f)} + {len(u_idx_b)} = {len(u_idx_f) + len(u_idx_b)} out"
				)

			plotax(ax_r, ax_c, 
				(bias[u_idx_f], bias[u_idx_b]),
				(r[u_idx_f], r[u_idx_b]),
				(cu[u_idx_f], cu[u_idx_b]),
				label=f"{u} {suffix}"
			)

	for ax in ax_r:
		ax.set_xlabel(r"Bias ($V$)")
		ax.set_ylabel(r"Sheet resistance ($\Omega/sq$)")
		ax.legend(loc=2)
	
	for ax in ax_c:
		ax.set_xlabel(r"Bias ($V$)")
		ax.set_ylabel(r"Conductance ($e^2 / h$)")
		ax.legend(loc=1)
	
	ax_r[0].set_title("Forward bias sweep")
	ax_c[0].set_title("Forward bias sweep")
	ax_r[1].set_title("Reverse bias sweep")
	ax_c[1].set_title("Reverse bias sweep")
	
	return (bias[idx_f], r[idx_f], c[idx_f]), (bias[idx_b], r[idx_b], c[idx_b]), (ax_r, ax_c)

def fullmob(f, b, ax, cox, lim=((-100, 100), (-100, 100))):
	mob_f, mob_b = None, None

	if lim[0] is not None:
		idx_f = np.where((f[0] > lim[0][0]) & (f[0] < lim[0][1]))[0]
		if len(idx_f) > 1:
			mob_f = fit_to_units(*mobility(cox, f[0][idx_f], f[2][idx_f]))
			ax[1][0].plot(f[0], mob_f[3]*f[0] + mob_f[4], label="Mobility fit")
			ax[1][0].legend(loc=1)
			print(f"Mobility (forward) = {mob_f[0]*1e4:.0f} cm²V⁻¹s⁻¹, r²={mob_f[1]**2:.4f}, std.err.={mob_f[2]:.4E}")

	if lim[1] is not None:
		idx_b = np.where((b[0] > lim[1][0]) & (b[0] < lim[1][1]))[0]
		if len(idx_b) > 1:
			mob_b = fit_to_units(*mobility(cox, b[0][idx_b], b[2][idx_b]))
			ax[1][1].plot(b[0], mob_b[3]*b[0] + mob_b[4], label="Mobility fit")
			ax[1][1].legend(loc=1)
			print(f"Mobility (reverse) = {mob_b[0]*1e4:.0f} cm²V⁻¹s⁻¹, r²={mob_b[1]**2:.4f}, std.err.={mob_b[2]:.4E}")

	return mob_f, mob_b

def contact_resistance(ds, r_sample, r_cur=10000):
	# Compute the current through the supply resistor by means of the measured voltage drop across it
	I = np.array([i.vi[0] / r_cur for i in ds.dataset])

	# Compute the voltage drop over the contacts + sample
	dU = np.array([i.i_set * r_cur - i.vi[0] for i in ds.dataset])

	# Solve Ohm's law, where R = 2R_c + R_s
	Rc = 0.5*(dU / I - r_sample)

	return Rc

class Sweep:
	# I assume that once the dataset is in this class, it will not be changed anymore
	def __init__(self, ds, vd=None, phase_scale=180):
		self.ds = ds
		self._vd = vd
		self.phase_scale = phase_scale
		self.gate_shift = True
		#if self.ds[0].type is not DatapointType.VDP:
		#	raise Exception(f"Provided dataset MUST be {DatapointType.VDP}, got {self.ds.type}")
		if len(np.unique([i.get_meta("ident") for i in self.ds])) > 1:
			raise Exception(f"More than one unique ident, this is not supported for `Sweep`")

	@cached_property
	def _bias(self):
		return np.array([i.bias for i in self.ds])

	@cached_property
	def _bias_err(self):
		# Note that these errors only work if we actually measure voltage at vi.
		return np.array([np.max(i.v_set - i.vi[0], 0) for i in self.ds])
	
	@cached_property
	def _res(self):
		return np.array([i.r[0] for i in self.ds])

	@cached_property
	def _res_err(self):
		return np.array([i.r[1] for i in self.ds])

	@cached_property
	def _phase_weight(self):
		p = 1 - np.abs([i.phase for i in self.ds]) / self.phase_scale
		p[p < 0.1] = 0.1
		return p

	@cached_property
	def _idx_f(self):
		return np.where(np.roll(self._bias, -1) > self._bias)[0]

	@cached_property
	def _idx_b(self):
		return np.where(np.roll(self._bias, -1) < self._bias)[0]

	@cached_property
	def dirac_voltage(self):
		if self._vd is not None:
			return self._vd

		vd = [None, None]
		if self.has_forward:
			idx = np.argmax(self._res[self._idx_f])
			vd[0] = self._bias[self._idx_f[idx]]
		
		if self.has_backward:
			idx = np.argmax(self._res[self._idx_b])
			vd[1] = self._bias[self._idx_b[idx]]

		return vd

	@cached_property
	def gate(self):
		vd = self.dirac_voltage

		gate = [None, None]
		if self.has_forward:
			gate[0] = self._bias[self._idx_f]
			if self.gate_shift:
				gate[0] -= vd[0]

		if self.has_backward:
			gate[1] = self._bias[self._idx_b]
			if self.gate_shift:
				gate[1] -= vd[1]

		return gate

	@property
	def has_forward(self):
		return len(self._idx_f) > 1
	
	@property
	def has_backward(self):
		return len(self._idx_b) > 1

	def forward_resistance(self):
		if not self.has_forward:
			return None

		return self.gate[0], self._res[self._idx_f], self._res_err[self._idx_f]

	def backward_resistance(self):
		if not self.has_backward:
			return None

		return self.gate[1], self._res[self._idx_b], self._res_err[self._idx_b]

	def forward_conductance(self, alt_units=False):
		if not self.has_forward:
			return None

		c = 1/self._res[self._idx_f]
		c_err = c**2 * self._res_err[self._idx_f]
		if alt_units:
			c = c * const.h / const.e**2
			c_err = c_err * const.h / const.e**2

		return self.gate[0], c, c_err

	def backward_conductance(self, alt_units=False):
		if not self.has_backward:
			return None

		c = 1/self._res[self._idx_b]
		c_err = c**2 * self._res_err[self._idx_b]
		if alt_units:
			c = c * const.h / const.e**2
			c_err = c_err * const.h / const.e**2

		return self.gate[1], c, c_err

	@lru_cache
	def mobilities(self, cox, exclusion_voltage=5, fit_limit=1000):
		def gen_sub():
			return {"hole": None, "electron": None}

		res = {
			"forward": None,
			"backward": None
		}

		if self.has_forward:
			sub = gen_sub()

			# Bias goes from negative to positive, so low bias index is hole
			bias, c, err = self.forward_conductance()

			if len(bias[bias < -exclusion_voltage]) > 2:
				# We have a few points to compute hole mobilities with
				bounds = (bias < -exclusion_voltage) & (bias > -fit_limit)

				mob = fit_to_units(*mobility(
					cox,
					bias[bounds],
					c[bounds],
					strat="curve_fit",
					err=err[bounds]
				))
				sub["hole"] = {
					"mobility": mob[0],
					"r_sq": mob[1]**2,
					"std_err": mob[2],
					"fit_param": {
						"alt_units": True,
						"slope": mob[3],
						"intercept": mob[4]
					}
				}

			if len(bias[bias > exclusion_voltage]) > 2:
				# We have a few points to compute electron mobilities with
				bounds = (bias > exclusion_voltage) & (bias < fit_limit)

				mob = fit_to_units(*mobility(
					cox,
					bias[bounds],
					c[bounds],
					strat="curve_fit",
					err=err[bounds]
				))
				sub["electron"] = {
					"mobility": mob[0],
					"r_sq": mob[1]**2,
					"std_err": mob[2],
					"fit_param": {
						"alt_units": True,
						"slope": mob[3],
						"intercept": mob[4]
					}
				}

			res["forward"] = sub

		if self.has_backward:
			sub = gen_sub()

			# Bias goes from negative to positive, so low bias index is electron
			bias, c, err = self.backward_conductance()

			if len(bias[bias < -exclusion_voltage]) > 2:
				# We have a few points to compute electron mobilities with
				bounds = (bias < -exclusion_voltage) & (bias > -fit_limit)

				mob = fit_to_units(*mobility(
					cox,
					bias[bounds],
					c[bounds],
					strat="curve_fit",
					err=err[bounds]
				))
				sub["hole"] = {
					"mobility": mob[0],
					"r_sq": mob[1]**2,
					"std_err": mob[2],
					"fit_param": {
						"alt_units": True,
						"slope": mob[3],
						"intercept": mob[4]
					}
				}

			if len(bias[bias > exclusion_voltage]) > 2:
				# We have a few points to compute hole mobilities with
				bounds = (bias > exclusion_voltage) & (bias < fit_limit)

				mob = fit_to_units(*mobility(
					cox,
					bias[bounds],
					c[bounds],
					strat="curve_fit",
					err=err[bounds]
				))
				sub["electron"] = {
					"mobility": mob[0],
					"r_sq": mob[1]**2,
					"std_err": mob[2],
					"fit_param": {
						"alt_units": True,
						"slope": mob[3],
						"intercept": mob[4]
					}
				}

			res["backward"] = sub

		return res

	def stats(self, cox, exclusion_voltage=5, fit_limit=1000, output=True, plot=True, plot_split=True, do_alpha=True, \
		figsize=(16,8), unshift=False):
		def plotfluff(ax, title):
			if self._vd == [0, 0] or unshift:
				ax.set_xlabel(r"Gate voltage ($V$)")
			else:
				ax.set_xlabel(r"Gate voltage ($V-V_{dirac}$)")
			ax.set_ylabel(r"Conductance ($e^2/h$)")
			ax.set_title(title)
			ax.legend()

		mob = self.mobilities(cox, exclusion_voltage=exclusion_voltage, fit_limit=fit_limit)
		c_f = self.forward_conductance(alt_units=True)
		c_b = self.backward_conductance(alt_units=True)

		if output:
			title = (f"Stats for {self.ds[0].get_meta('ident')} - "
				f"{self.ds[0]._header['id']} ({self.ds[0].probe_type}) - "
				f"n={len(self.ds)}")
			print(f"{title:-^80}")

			# Mobilities
			print("Mobility")
			if self.has_forward:
				if mob["forward"]["hole"] is not None:
					cur = mob["forward"]["hole"]
					print(f"{'': <15}μₕ = {cur['mobility']*1E4: >5.2f} ± {cur['std_err']*1E4:.2f} cm²V⁻¹s⁻¹ (forward)")
				if mob["forward"]["electron"] is not None:
					cur = mob["forward"]["electron"]
					print(f"{'': <15}μₑ = {cur['mobility']*1E4: >5.2f} ± {cur['std_err']*1E4:.2f} cm²V⁻¹s⁻¹ (forward)")
			if self.has_backward:
				if mob["backward"]["hole"] is not None:
					cur = mob["backward"]["hole"]
					print(f"{'': <15}μₕ = {cur['mobility']*1E4: >5.2f} ± {cur['std_err']*1E4:.2f} cm²V⁻¹s⁻¹ (reverse)")
				if mob["backward"]["electron"] is not None:
					cur = mob["backward"]["electron"]
					print(f"{'': <15}μₑ = {cur['mobility']*1E4: >5.2f} ± {cur['std_err']*1E4:.2f} cm²V⁻¹s⁻¹ (reverse)")

			# Conductivities
			print("\nMinimum conductance")
			if self.has_forward:
				print(f"{'': <15}σ = {np.min(c_f[1]):.2f} e²/h (forward)")
			if self.has_backward:
				print(f"{'': <15}σ = {np.min(c_b[1]):.2f} e²/h (reverse)")

			# Dirac voltages
			print("\nDirac voltage")
			if self.has_forward:
				print(f"{'': <15}Vd = {self.dirac_voltage[0]:.2f} V (forward)")
			if self.has_backward:
				print(f"{'': <15}Vd = {self.dirac_voltage[1]:.2f} V (reverse)")

		if plot:
			if plot_split:
				plt_idx = (1, 2)
			else:
				plt_idx = (1, 1)

			fig, ax = plt.subplots(*plt_idx, figsize=figsize)
			self.fig = fig

			if plot_split:
				ax_f, ax_b = ax
			else:
				ax_f = ax
				ax_b = ax

			if self.has_forward:
				v = self.dirac_voltage[0] if unshift else 0
				if do_alpha:
					ax_f.scatter(
						c_f[0]+v, c_f[1], s=3, \
						color="black", alpha=self._phase_weight[self._idx_f], label="Forward"
					)
				else:
					ax_f.errorbar(
						c_f[0]+v, c_f[1], xerr=self._bias_err[self._idx_f], yerr=c_f[2], ms=3, \
						color="black", label="Forward", fmt='.'
					)

				if mob["forward"]["hole"] is not None:
					cur = mob["forward"]["hole"]
					x = np.array([np.max([np.min(c_f[0]), -fit_limit]), -exclusion_voltage])
					ax_f.plot(x+v, cur["fit_param"]["slope"] * x + cur["fit_param"]["intercept"], label="μₕ (forward)")
				if mob["forward"]["electron"] is not None:
					cur = mob["forward"]["electron"]
					x = np.array([exclusion_voltage, np.min([np.max(c_f[0]), fit_limit])])
					ax_f.plot(x+v, cur["fit_param"]["slope"] * x + cur["fit_param"]["intercept"], label="μₑ (forward)")

				plotfluff(ax_f, "Forward sweep")

			if self.has_backward:
				v = self.dirac_voltage[1] if unshift else 0
				col = "grey" if unshift else "black"
				if do_alpha:
					ax_b.scatter(
						c_b[0]+v, c_b[1], s=3, \
						color="black", alpha=self._phase_weight[self._idx_b], label="Reverse"
					)
				else:
					ax_b.errorbar(
						c_b[0]+v, c_b[1], xerr=self._bias_err[self._idx_b], yerr=c_b[2], ms=3, \
						color=col, label="Reverse", fmt='.'
					)

				if mob["backward"]["hole"] is not None:
					cur = mob["backward"]["hole"]
					x = np.array([np.max([np.min(c_b[0]), -fit_limit]), -exclusion_voltage])
					ax_b.plot(x+v, cur["fit_param"]["slope"] * x + cur["fit_param"]["intercept"], label="μₕ (reverse)")
				if mob["backward"]["electron"] is not None:
					cur = mob["backward"]["electron"]
					x = np.array([exclusion_voltage, np.min([np.max(c_b[0]), fit_limit])])
					ax_b.plot(x+v, cur["fit_param"]["slope"] * x + cur["fit_param"]["intercept"], label="μₑ (reverse)")

				plotfluff(ax_b, "Reverse sweep" if plot_split else "")

			plt.show()

	def export(self):
		self.fig.savefig(f"{self.ds[0].get_meta('ident')}.svg", format="svg")
