import re
from io import StringIO
from functools import cached_property
from enum import Enum
from warnings import warn
import numpy as np

from tools import read, pin

class Data:
	def __init__(self, filename, raw):
		# We need to do this, as setting these outside of the constructor means that
		# they store the same pointer ensuring that setting these for one instance
		# sets them for all others too.
		self.header = { \
			"file": "", \
			"type": "", \
			"probe": "", \
			"id": "", \
			"desc": "", \
			"steps": ""
		}
		self.data = np.array([])

		headers = (("type", "Measurement type"), ("probe", "probe type"), ("desc", "user info"), ("id", "probe ID"), ("steps", "steps"))
		self.header["file"] = filename
		for val in headers:
			self.header[val[0]] = self._getValueFromRaw(val[1], raw)

		raw_data = self._getSegmentFromRaw("Data", raw)

		# Drop the datetime and correct columns
		# TODO: Don´t drop the correct column, although this doesn´t matter too much as we can deduce
		# the information from this column from other columns.
		usecols = [x for x in range(27) if x not in (23, 25)]

		# Skip the header row. Additionally, force the output to be a 2D array. Otherwise, is we have only one
		# measurement in a data file, when iterating over this data we iterate per entry instead of per row
		self.data = np.loadtxt(StringIO(raw_data), delimiter=",", skiprows=1, usecols=usecols, ndmin=2)

	def _getValueFromRaw(self, needle, haystack):
		regex = re.search(fr"^{needle}=(.*)$", haystack, re.MULTILINE)
		return regex.group(1)

	def _getSegmentFromRaw(self, needle, haystack, replace_tab=True):
		regex = re.search(fr"\[{needle}\](.*)", haystack, re.MULTILINE | re.DOTALL)
		res = regex.group(1)
		if not replace_tab:
			return res

		res = re.sub(r" +", "", res)
		res = re.sub(r"\t", ",", res)
		res = re.sub(r",\n", "\n", res)

		# Due to the regex substitutions we are left with
		# a newline at the start of the string. Drop that
		return res[1:]

class DatapointType(Enum):
	NONE = 0
	NORMAL = 1
	VDP = 2

class Datapoint:
	Z_KEITH = 1E6
	Z_LOCKIN = 10E6

	RES_DAT = {
		"10k": (9.942943e3, 43.63123e-3),
		"20k": (19.84736e3, 399.5066e-3),
		"100k": (109.1314e3, 725.3434e-3),
		"1M": (996.4879e3, 6.856803),
		"5M": (4.963704e6, 1.130532e3)	# This std.dev. seems rather high. Maybe remeasure?
	}

	# Data repository: Some older data refers to these resistances a little differently
	# So, create them seperately
	RES_DAT["110k"] = RES_DAT["100k"]

	def __init__(self, row, probe_type="narrow", header=None, meta=None):
		self.type = DatapointType.NORMAL
		self.probe_type = probe_type

		self.pids = row[:4].astype(int)
		self.i_set = row[4]
		self.vi = row[5:7]
		self.v = row[7:9]
		self.r = row[9:11]
		self.phase = row[11]
		self.field = row[21]
		self.freq = row[-2]
		self.offset = row[-1]
		self.avg_no = row[15]
		self._dist = None
		self._meta = {}
		self._header = header

		if meta is not None and header is not None:
			for key, rule, cast in meta:
				regex = re.search(rule, header["desc"])
				regex2 = re.search(rule, header["file"])

				# Check if we want to match against the description or file name
				if not regex and regex2:
					# Desc no match, filename match
					regex = regex2
				elif not regex and not regex2:
					# No matches
					warn(f"No meta matches found for {key} in {header['file']}")
					continue
				else:
					# Else, we match against the description. Should we wish to match against the filename instead,
					# we should use a stricter rule
					pass
				
				self._meta[key] = cast(regex.group(1))

			self.fix_all_fucked()

	def __str__(self):
		return (f"<Datapoint> PT={self.probe_type}; PINS={self.pids}; FILE={self._header['file']};"
			f"DATA=[I={self.i_set:.2e}; Vi={self.vi[0]:.2e}; V={self.v[0]:.2e}; R={self.r[0]:.2e};"
			f"Phase={self.phase:.2f}; Bias={self.bias:.2e}; f={self.freq}")

	# Lets not cache this property in case we call it first and change it afterwards
	@property
	def dist(self):
		if self._dist is not None:
			return self._dist

		sort = np.sort(self.pids)
		return np.mean([
			pin.PinDist.compute_distance(
				self.probe_type,
				int(sort[i]),
				int(sort[i+1])
			)
			for i in range(len(sort) - 1)
		])
	
	@property
	def pin_str(self):
		return pin.PinDist.ids_to_str(self.pids)
	
	@property
	def bias(self):
		# Due to the way the system was wired, positive field corresponds to negative gate bias
		# Additionally, convert from the input kA/m units to V, via a conversion factor for the
		# Kepco
		# See `field-bias-rainbow.ipynb` for the fit
		return -5.3519E-4 * self.field - 3.2883E-2

	@property
	def r_cur(self):
		# If we used a different current resistor, we can call this function to compute it, as well as the standard
		# deviation of the measured resistance over 128 points (slow rate with Keithley 2700 2point measurements).
		if self.has_meta("qty") and self.has_meta("res"):
			return Datapoint.RES_DAT[f"{self.get_meta('res')}{self.get_meta('qty')}"]
		else:
			return Datapoint.RES_DAT["10k"]

	@property
	def v_set(self):
		# Compute the output voltage set on the lock-in. This isn't exactly the value set on the display, as the
		# resistances r_cur are measured more accurately then their nominal values.  We assume this property is only
		# called after `recompute_resistance` has been ran.
		return self.i_set * self.r_cur[0]
	
	@dist.setter
	def dist(self, new_dist):
		self._dist = new_dist

	# We need to pass either a list or a list of lists here. We expect this to be
	# (object key, value to test against elem's first entry, module 'operator' math operation)
	# Returns: True if object passes filter tests, False if it doesn´t and should be filtered
	def filter(self, attr, with_self=False):
		def test_attr(key, val, op):
			stored = getattr(self, key)
			try:
				# Very dirty check to see if this is a list or a single number
				len(stored)
				to_test = stored[0]

				# If we want to check pin IDs, we'll need all of them
				# If we are checking for a string, well, a string is a list of characters under the hood
				if key == "pids" or key == "probe_type":
					to_test = stored
			except:
				to_test = stored
			finally:
				if with_self:
					return op(to_test, val, self)
				else:
					return op(to_test, val)
			
		if type(attr[0]) is tuple or type(attr[0]) is list:
			res = [test_attr(*i) for i in attr]
			return False not in res
		else:
			return test_attr(*attr)

	def has_meta(self, key=None):
		if key is not None:
			return key in self._meta
		
		return len(self._meta.items()) != 0

	def set_meta(self, key, val):
		self._meta[key] = val

	def get_meta(self, key):
		return self._meta[key]

	# Same rules as filter() apply here, apart from that we only check for equality
	def filter_meta(self, key, val):
		if not self.has_meta(key):
			warn(f"Filter_meta {key=} not found")
			return False

		return self.get_meta(key) == val

	# Recompute the resistances based on a different current resistor
	def recompute_resistance(self, *args):
		# Args is unused
		# Note, the Keithley has a 1Mohm impedance in AC mode. Thus, we are measuring the voltage over two
		# parallel resistors. In DC mode, this doesn't need to be compensated for.
		if self.has_meta("type") and self.get_meta("type") == "keithdc":
			r = self.r_cur[0]
		else:
			r = self.r_cur[0]*Datapoint.Z_KEITH / (self.r_cur[0] + Datapoint.Z_KEITH)

		# If we set the keithley in DC Current mode, vi holds said current
		if self.has_meta("type") and self.get_meta("type") == "keithcur":
			i = self.vi[0]
		else:
			i = self.vi[0] / r

		if i == 0.:
			warn(f"Zero current, resistance computation invalid for {self}")

		dist_pins = (
			(self.pids[2], self.pids[0]),
			(self.pids[1], self.pids[2]),
			(self.pids[3], self.pids[0]),
			(self.pids[1], self.pids[3])
		)

		s = [pin.PinDist.compute_distance(self.probe_type, *pins) for pins in dist_pins]

		fac = np.log((s[1] * s[2]) / (s[0] * s[3]))

		# Next, we also need to deal with the fact that the lock-in has a 10Mohm impedance per input.
		# This should in theory be of little to no influence.
		rtot = self.v[0] / i
		self.r[0] = rtot * Datapoint.Z_LOCKIN / (Datapoint.Z_LOCKIN - rtot)
		self.r[0] = 2*np.pi*self.r[0] / fac
		self.r[1] = None	# As we have recomputed stuff, the original error here is invalid
		self.i_set *= 10E3 / self.r_cur[0]

	# Fix a number of things that can be different from what we might expect, in order for stuff to work as we expect
	# it to.
	def fix_all_fucked(self):
		# Fix nano400 probe not being present in software
		if self.has_meta("probe"):
			p = self.get_meta("probe")
			self.probe_type = p
			self._header["probe"] = p

		# Recompute the resistances. In some occations, we don´t need to. However, at this point this is the exception,
		# not the rule, so we might as well just do it anyway. The only thing we lose are the errors, but we can
		# add those computations if required. When doing VDP, we recompute all this stuff anyway
		self.recompute_resistance()

	@staticmethod
	def to_vdp(p1, p2, filter=None, with_self=False):
		# Check if we are inputting normal data points
		if not (p1.type is DatapointType.NORMAL and p2.type is DatapointType.NORMAL):
			raise ValueError(
				f"Both arguments should be normal data points, not NONE or VDP. Got {p1.type} and {p2.type}")

		# Check if we need to filter the inputs
		if filter is not None:
			fil = [i.filter(filter, with_self) for i in (p1, p2)]

			# This combo needs to be filtered out, hence we return a None object
			if False in fil:
				return None

		# Compute the probe pair types for use in vdp correction
		types = [pin.PairType.get_from_pins(i.pids) for i in (p1, p2)]

		# Compute measured resistances
		# If we used the resistor switchboard, we must set the current resistor accordingly
		r_cur, r_cur_err = p1.r_cur
		# Note, the Keithley has a 1Mohm impedance in AC mode. Thus, we are measuring the voltage over two
		# parallel resistors. In DC mode, this can be ignored.
		if (p1.has_meta("type") and p1.get_meta("type") == "keithdc") \
			and (p2.has_meta("type") and p2.get_meta("type") == "keithdc"):
			# Do nothing, the inversion of this if statement is just more annoying
			pass
		else:
			r_cur = r_cur*Datapoint.Z_KEITH / (r_cur + Datapoint.Z_KEITH)
			r_cur_err = Datapoint.Z_KEITH**2 / (r_cur + Datapoint.Z_KEITH)**2 * r_cur_err

		# Next, we also need to deal with the fact that the lock-in has a 10Mohm impedance per input.
		# This should in theory be of little to no influence. If we used current measurements, we compute
		# This a little differently
		if (p1.has_meta("type") and p1.get_meta("type") == "keithcur") \
			and (p2.has_meta("type") and p2.get_meta("type") == "keithcur"):
			r_tot = np.array([
				p1.v[0] / p1.vi[0],
				p2.v[0] / p2.vi[0],
			])

			r_tot_err = np.sqrt(
				(1 / np.array([p1.vi[0], p2.vi[0]]) * np.array([p1.v[1], p2.v[1]]))**2
				+ (-1 * np.array([
					p1.v[0] / p1.vi[0]**2,
					p2.v[0] / p2.vi[0]**2,
				]) * np.array([p1.vi[1], p2.vi[1]]))**2
			)
		else:
			r_tot = r_cur * np.array([
				p1.v[0] / p1.vi[0],
				p2.v[0] / p2.vi[0],
			])
			r_tot_err = np.sqrt(
				(np.array([
					p1.v[0] / p1.vi[0],
					p2.v[0] / p2.vi[0],
				]) * r_cur_err)**2
				+ (r_cur / np.array([p1.vi[0], p2.vi[0]]) * np.array([p1.v[1], p2.v[1]]))**2
				+ (-r_cur * np.array([
					p1.v[0] / p1.vi[0]**2,
					p2.v[0] / p2.vi[0]**2,
				]) * np.array([p1.vi[1], p2.vi[1]]))**2
			)
		r_meas = 2*np.pi * r_tot * Datapoint.Z_LOCKIN / (Datapoint.Z_LOCKIN - r_tot)
		r_meas_err = 2*np.pi*Datapoint.Z_LOCKIN**2 / (Datapoint.Z_LOCKIN - r_tot)**2 * r_tot_err

		# Normalise to AC pair type combo
		status, r_norm, r_norm_err = pin.Correction.normalise(r_meas, r_meas_err, types)
		if not status:
			# Data repository: Instead of erroring out, ignore any measurements that have no
			# (valid) vdp pairs pairs
			#raise ArithmeticError(f"Could not normalise pins to AC from {types}")
			return None

		# Compute vdp resistance
		cor, cor_err = pin.Correction.van_der_pauw(r_norm, r_norm_err, max_iter=50)

		# No bisection result could be found, so filter out these points
		if cor is False:
			return None

		# Initialise new datapoint and set relevant values
		# Errors are computed using error propagation for averaging
		err_avg = lambda x, y: np.sqrt((x/2)**2 + (y/2)**2)

		ret = Datapoint(np.zeros(30), p1.probe_type)
		ret.type = DatapointType.VDP
		ret.probe_type = p1.probe_type

		ret.pids = p1.pids
		# We do a bit of hacky shit here when having current measurements
		if (p1.has_meta("type") and p1.get_meta("type") == "keithcur") \
			and (p2.has_meta("type") and p2.get_meta("type") == "keithcur"):
			# I am about 100% certain that this is going to fuck me one way or another.
			ret.i_set = p2.vi
			ret.vi = p1.vi
		else:
			ret.i_set = p1.i_set
			ret.vi = np.array([(p1.vi[0] + p2.vi[0])/2, err_avg(p1.vi[1], p2.vi[1])])
		ret.v = np.array([(p1.v[0] + p2.v[0])/2, err_avg(p1.v[1], p2.v[1])])
		ret.r = np.array([cor, cor_err])
		ret.phase = np.mean([p1.phase, p2.phase])
		ret.field = p1.field
		ret.freq = p1.freq
		ret.offset = None
		ret.avg_no = p1.avg_no
		ret.dist = np.mean([p1.dist, p2.dist])	# TODO: Verify that this is correct
		ret._meta = p1._meta
		ret._header = p1._header

		return ret

class DatapointSet:
	def __init__(self, dataset=None, filtered=False):
		# We need to do this, as setting dataset=[] in the arguments means that
		# every separate instance of DatapointSet references the same list through its
		# pointer, meaning that setting the dataset of one instances sets the same for
		# all other instances
		if dataset is not None:
			self.dataset = dataset
		else:
			self.dataset = []

		self.filtered = filtered

	def __len__(self):
		return len(self.dataset)

	def __iter__(self):
		return self.dataset.__iter__()

	def __getitem__(self, index):
		return self.dataset[index]

	def fromFileList(self, files, meta=None):
		for file in files:
			d = Data(*read.readFile(file))

			self.dataset += [Datapoint(row, d.header["probe"], d.header, meta) for row in d.data]

	def filter(self, attr, new=True, with_self=False):
		res = [i.filter(attr, with_self) for i in self.dataset]
		# Note: We cannot use `if keep is True` here, as type(keep) = numpy.bool_ != type(True)
		ds = [dp for dp, keep in zip(self.dataset, res) if keep]

		if new:
			return DatapointSet(ds, True)
		else:
			self.dataset = ds
			self.filtered = True

	def filter_meta(self, *args, new=True):
		# We do not need to set the filtered flag here, as meta is header scope, not individual
		# measurement scope. Hence, we cannot break up vdp pairs using meta filtering
		res = [i.filter_meta(*args) for i in self.dataset]
		ds = [dp for dp, keep in zip(self.dataset, res) if keep]

		if new:
			return DatapointSet(ds)
		else:
			self.dataset = ds

	def van_der_pauw(self, filter, new=True, with_self=False):
		if self.filtered:
			raise Exception("This set has already been filtered. This will break van der Pauw as a single measurement that makes up the pair may have been filtered out")

		res = [Datapoint.to_vdp(i, j, filter, with_self) for i, j in zip(self.dataset[:-1:2], self.dataset[1::2])]
		ds = [dp for dp in res if dp is not None]

		if new:
			return DatapointSet(ds, True)
		else:
			self.dataset = ds
			self.filtered = True

	def is_vdp_able(self):
		# Check if points have already been VDP'd
		vdp = [i.type == DatapointType.VDP for i in self.dataset]

		if any(vdp):
			# Some points have already had VPD applied to them. Figure it out yourself
			return False

		# Next, check if we have correction values set. We could also check consecutive pairs, but this might become
		# a bit of a problem when we have a single entry that isn´t vdpable.
		check = lambda x: x.offset == 1 or x.offset == -1
		vdp = [check(i) for i in self.dataset]

		return vdp

