import numpy as np
from scipy.optimize import bisect

class PinDist:
	# Distances in um
	db = {
		"wide": (60, 24, 15, 6, 3, 3, 3, 3, 9, 12, 39),
		"standard": (9.6, 9.3, 4.5, 3, 1.5, 1.5, 1.5, 1.5, 5, 5.7, 12.4),
		"narrow": (4.5, 3, 1.5, 1.5, 1.5, 1.5, 1.75, 2, 2.25, 2.5, 2.75),
		"nano": (2.2, 1.5, 0.75, 0.75, 0.75, 0.75, 1.4, 1.0, 1.9, 3.1, 3.4),	#TODO: Verify these lengths
		"nano400": (2.0, 1.4, 0.6, 0.4, 0.4, 0.4, 0.4, 0.8, 1.2, 1.6, 2.4)
	}

	@staticmethod
	def compute_distance(probe_type, *ids):
		if len(ids) != 2:
			raise ValueError(f"ids should contain 2 pin IDs, got {len(ids)}")

		if not probe_type in PinDist.db:
			raise ValueError(f"Probe type {probe_type} is unknown")

		if ids[1] < ids[0]:
			ids = ids[::-1]

		cur_probe = np.array(PinDist.db[probe_type])
		idx = [i - 1 for i in ids]
		idx = [i for i in range(*idx)]

		return sum(cur_probe[idx])

	@staticmethod
	def ids_to_str(ids):
		return "-".join(ids.astype(int).astype(str))


class PairType:
	NONE = 0b00
	A = 0b01
	B = 0b10
	C = 0b11
	
	pt = NONE
	neg = False

	def __init__(self, pt=NONE, neg=False):
		self.pt = pt
		self.neg = neg

	def __repr__(self):
		txt = "None"
		if self.pt == PairType.A:
			txt = "A"
		elif self.pt == PairType.B:
			txt = "B"
		elif self.pt == PairType.C:
			txt = "C"

		return f"{'-' if self.neg else ''}{txt}"

	def __str__(self):
		return self.__repr__()

	def __eq__(self, other):
		if(isinstance(other, PairType)):
			return self.pt == other.pt and self.neg == other.neg
		
		return self.pt == other

	def __gt__(self, other):
		if(isinstance(other, PairType)):
			return self.pt > other.pt

		return self.pt > other

	def __lt__(self, other):
		if(isinstance(other, PairType)):
			return self.pt < other.pt

		return self.pt < other

	@staticmethod
	def get_from_pins(pins):
		if len(pins) != 4:
			raise ValueError(f"pins should contain 4 pin IDs, got {len(pins)}")

		def increasing(l):
			return all(a > b for a, b in zip(l, l[1:]))

		def decreasing(l):
			return all(a < b for a, b in zip(l, l[1:]))

		def monotonic(l):
			return increasing(l) or decreasing(l)

		A = (pins[0], pins[2], pins[3], pins[1])
		B = (pins[0], pins[2], pins[1], pins[3])
		C = (pins[0], pins[1], pins[3], pins[2])

		if monotonic(A):
			return PairType(PairType.A)
		elif monotonic(B):
			return PairType(PairType.B)
		elif monotonic(C):
			return PairType(PairType.C)

		return PairType()

class Correction:
	@staticmethod
	def pairs_to_int(*args):
		if len(args) != 2:
			raise ValueError(f"Expected 2 arguments, got {len(args)}")

		res = args[1].neg << 5 | args[1].pt << 3 | args[0].neg << 2 | args[0].pt
		
		return res & 0b111111

	@staticmethod
	def normalise(r, r_err, pairs):
		ret = (False, None, None)
		sum_err = lambda a, b: np.sqrt(a**2 + b**2)

		# Check tuple length, this should stop script exec
		if len(r) != 2 or len(r_err) != 2:
			raise ValueError(f"r (r_err) should be length 2, got {len(r)} ({len(r_err)})")

		# Check tuple length, this should stop script exec
		if len(pairs) != 2:
			raise ValueError(f"pairs should be length 2, got {len(pairs)}")

		# Check equality. From here, checks should not stop script exec, as some parameter input choices
		# may be intentional
		if pairs[0] == pairs[1]:
			return ret

		# Apply negations to resistances, if applicable
		r = [-1*val if pairs[i].neg else val for i, val in enumerate(r)]
		r_err = [-1*val if pairs[i].neg else val for i, val in enumerate(r_err)]

		# Flip the data if required
		if pairs[0] > pairs[1]:
			r = r[::-1]
			r_err = r_err[::-1]
			pairs = pairs[::-1]

		# Apply conversion
		if pairs[0] == PairType.A and pairs[1] == PairType.B:
			ret = (True, (r[0], r[0] - r[1]), (r_err[0], sum_err(*r_err)))
		elif pairs[0] == PairType.A and pairs[1] == PairType.C:
			ret = (True, tuple(r), tuple(r_err))
		elif pairs[0] == PairType.B and pairs[1] == PairType.C:
			ret = (True, (r[0] + r[1], r[1]), (sum_err(*r_err), r_err[1]))

		# All other conversions are invalid, so we are done.
		# Return whatever the result was
		return ret

	@staticmethod
	def van_der_pauw(r, r_err, epsilon=1e-10, max_iter=100):
		# This function kills the execution of no appropriate bounds are found
		def check_and_kill(iter, rl, rh, res):
			if iter < max_iter:
				return

			raise Exception(f"No bounds found: {rl=}, {rh=}, {res=}")

		if len(r) != 2 or len(r_err) != 2:
			raise ValueError(f"r (r_err) should be length 2, got {len(r)} ({len(r_err)})")

		# The reason that this is the equation we solve, is because we put in
		# 2*pi*V/I. Then this follows eq. (4.5) of miccoli2015
		func = lambda rs: np.exp(-r[0]/rs) + np.exp(-r[1]/rs) - 1

		# Compute appropriate bounds
		rb = sum(r)/2
		res = func(rb)
		biter = 0

		if res > 0:
			rh = rb
			rl = rb/2
			while func(rl) > 0 and not np.isinf(func(rl)):
				rl /= 2
				biter += 1
				check_and_kill(biter, rl, rh, func(rl))
		else:
			rl = rb
			rh = rb*2
			while func(rh) < 0 and not np.isinf(func(rh)):
				rh *= 2
				biter += 1
				check_and_kill(biter, rl, rh, func(rh))

		# If we encounter an overflow in determining the bounds, we return as if we have no result
		if np.isinf(func(rl)) or np.isinf(func(rh)):
			return False, None

		# Compute the position corrected resistance using a bisection method
		vdp = bisect(
			func,
			rl, rh,
			xtol=epsilon,
			maxiter=max_iter
		)

		# Compute the error in the correction based on implicit derivation of the bisection function
		denom = r[0] * np.exp(-r[0]/vdp) + r[1] * np.exp(-r[1]/vdp)
		dra = vdp * np.exp(-r[0]/vdp) / denom
		drb = vdp * np.exp(-r[1]/vdp) / denom

		vdp_err = np.sqrt((dra * r_err[0])**2 + (drb * r_err[1])**2)

		return vdp, vdp_err
