from functools import reduce

import numpy as np

from tools.data import DatapointSet

class Stat:
	def __init__(self, ds, expected_val=1.8):
		if not isinstance(ds, DatapointSet):
			raise ValueError("Provided argument is not a DatapointSet")

		self.id = ds[0]._header["id"]
		self.type = ds[0].probe_type

		self.optim = expected_val

		self.res = self._perform_analysis(ds)
		self.res_min = self._perform_analysis_min(ds)
		self.pin_ids = self._get_unique_pin_ids(ds)

	def _get_unique_pin_ids(self, ds):
		# We cannot unpack inside list comprehensions. Hence, we do this
		pids = []
		for i in ds.dataset:
			pids += list(i.pids)

		return np.unique(pids)

	def _perform_analysis(self, ds):
		pin_ids = self._get_unique_pin_ids(ds)
		#print(pin_ids, len(pin_ids))

		# [deviation from optimum, no. of points] for every pin number
		stats = {i:[0, 0] for i in pin_ids}

		# Compute the deviation for each pin
		for entry in ds.dataset:
			deviation = np.abs(entry.r[0] - self.optim)
			for pin in entry.pids:
				stats[pin][0] += deviation
				stats[pin][1] += 1

		# Find the pin with the maximum average deviation
		key = reduce(lambda idx, key: \
			idx if stats[idx][0]/stats[idx][1] > stats[key][0]/stats[key][1] else key, stats)

		worst = (key, *stats[key])

		# Recurse analysis with the remaining pins
		new_ds = ds.filter(("pids", key, lambda pids, key: key not in pids))
		if len(new_ds.dataset) > 0:
			# There are still enough combinations left for more analysis
			ana = self._perform_analysis(new_ds)

			# Analysis complete, return results
			ana.append(worst)
			return ana
		else:
			# No more combinations possible. This means we are left with 4 pins. Sort them by average deviation
			# Best pin first.
			vals = [stats[i][0]/stats[i][1] for i in stats]
			sorted_idx = np.argsort(vals)

			# Convert idx to keys
			keys = list(stats.keys())
			sorted_keys = [keys[i] for i in sorted_idx]

			# Return as list of tuples
			return [(i, *stats[i]) for i in sorted_keys]

	def _perform_analysis_min(self, ds):
		def compute(ds, stats):
			pin_ids = self._get_unique_pin_ids(ds)
			tmp = {i:[0, 0] for i in pin_ids}

			for entry in ds.dataset:
				dev = np.abs(entry.r[0] - self.optim)
				for pin in entry.pids:
					tmp[pin][0] += dev
					tmp[pin][1] += 1

			for pin in tmp:
				avg = tmp[pin][0] / tmp[pin][1]
				if avg < stats[pin]:
					stats[pin] = avg

			return stats

		# Here we keep track of the minimum average deviation
		pin_ids = self._get_unique_pin_ids(ds)
		stats = {i:1E99 for i in pin_ids}

		# First, drop no pins. Then, drop 1 pin at a time
		stats = compute(ds, stats)
		for i in pin_ids:
			new_ds = ds.filter(("pids", i, lambda pids, key: key not in pids))
			stats = compute(new_ds, stats)

		return stats

	def _compute_stats(self):
		res = []
		for pin in self.pin_ids:
			res += [(i[0], np.abs(i[1] / i[2]) * 100 / self.optim, i[1] / i[2]) for i in self.res if i[0] == pin]

		return res

	def print_stats(self):
		print(f"--- Stats for probe {self.id} ({self.type}) ---")
		print(f"|{'PID':>5}|{'Deviation':>16}|{'Average':>15}|")
		for i in self._compute_stats():
			print(f"|{i[0]:>5}|{i[1]:>15.2f}%|{i[2]:>15.2f}|")

	def print_stats_min(self):
		print(f"--- Min. stats for probe {self.id} ({self.type}) ---")
		print(f"|{'PID':>5}|{'Min Avg Dev':>15}|")
		for i in self.res_min:
			print(f"|{i:>5}|{self.res_min[i]:>15.2f}|")